/* $Header: /CVSROOT/filecache/filecache.c,v 1.2 2006/11/15 02:34:26 tino Exp $
 *
 * filecache: Create a cache entry for a filename
 * Copyright (C)2006 Valentin Hilbig, webmaster@scylla-charybdis.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * $Log: filecache.c,v $
 * Revision 1.2  2006/11/15 02:34:26  tino
 * Options -b -d -e and -z
 *
 * Revision 1.1  2006/11/13 04:44:18  tino
 * first checkin
 *
 */

#include "tino/filetool.h"
#include "tino/slist.h"
#include "tino/ex.h"
#include "tino/buf_line.h"
#include "tino/getopt.h"
#include "tino/dir.h"
#include "tino/str.h"

#include <unistd.h>

#include "filecache_version.h"

#if 0
static int	ignorecase, utf8;
#endif

struct filecache_user
  {
    TINO_BUF	namebuf, nameout;
    const char	*found;
  };

static char	codepage[256];
static char	codechar[256];

static char	in_termchar, out_termchar;
static int	null_termchar, blanks_termchar, maxdir, maxpath, readonly, readfiles, debug_mode;

static void
deb(const char *s, ...)
{
  va_list	list;

  if (!debug_mode)
    return;
  fprintf(stderr, "filecache debug: ");
  va_start(list, s);
  vfprintf(stderr, s, list);
  va_end(list);
  fprintf(stderr, "\n");
}

static void
out(TINO_BUF *buf)
{
  printf("%s%c", tino_buf_get_s(buf), (out_termchar ? out_termchar : null_termchar ? 0 : 10));
}

#define	FILECACHE_SPECIAL_CHAR	'$'

static int
mapchar(int i)
{
  switch (i)
    {
    case '#':
    case '%':
    case '+':
    case ',':
    case '-':
#if 0
    case '.':	/* Dots are directory special, see . and ..	*/
#endif
    case '@':
    case '_':
    case '~':
      return i;
    }
  if ((i>='0' && i<='9') || (i>='a' && i<='z'))
    return i;
  if (i>=32 && i<64)
    {
      i	&= 31;
      if (i<16)
	return 'a'+i;
      return 'a'+i-10;
    }
  i	&= 31;
  if (!i)
    return '0';
  if (i>26)
    return '0'+i-26;
  return 'a'+i-1;
}

static int
filecache_pathpart(const char *s, void *u)
{
  struct filecache_user	*user=(void *)u;
  size_t		len;

  if (!s || !*s || tino_strprefixcmp(tino_buf_get(&user->namebuf), s))
    return 0;
  len	= strlen(s);
  if (s[len-1]==FILECACHE_SPECIAL_CHAR && len<tino_buf_get_len(&user->namebuf))
    return 0;
  deb("found entry %s", s);
  return len;
}

static void
filecache_init(void)
{
  int	i;

  for (i=256; --i>=0; )
    {
      codepage[i]	= i/32;
      codechar[i]	= mapchar(i);
#if 0
      printf("%d%c\n", codepage[i], codechar[i]);
#endif
    }
}

static void
filecache_mangle(const char *name, struct filecache_user *user)
{
  int	last_cp;

  deb("mangling %s", name);
  tino_buf_reset(&user->namebuf);
  last_cp	= codepage['a'];
  while (*name)
    {
      unsigned char	c;
      int		cp, code;

      c		= *name++;
      c		&= 0xff;
      cp	= codepage[c];
      code	= codechar[c];
      if (cp!=last_cp)
	{
	  tino_buf_add_c(&user->namebuf, FILECACHE_SPECIAL_CHAR);
	  tino_buf_add_c(&user->namebuf, '0'+cp);
	  last_cp	= cp;
	}
      tino_buf_add_c(&user->namebuf, code);
    }
  /* end name on $ to differ from directory names
   *
   * Else we might hit a problem in creating a cache entry for 'ab'
   * where 'a' already exists (it's expected to be a directory then).
   *
   * This way 'a' is 'a$' and 'ab' is 'ab$'
   */
  tino_buf_add_c(&user->namebuf, FILECACHE_SPECIAL_CHAR);
  deb("mangled %s", tino_buf_get_s(&user->namebuf));
}

static void
filecache_advance(int found, struct filecache_user *user)
{
  tino_buf_add_c(&user->nameout, TINO_PATH_SEP_CHAR);
  tino_buf_add_n(&user->nameout, tino_buf_get(&user->namebuf), found);
  tino_buf_advance(&user->namebuf, found);
}

static void
filecache_new(struct filecache_user *user)
{
  int	max;

  max	= tino_buf_get_len(&user->namebuf);
  if (max<=maxpath)
    {
      filecache_advance(max, user);
      return;
    }
  max	= maxpath;
  if (tino_buf_get_s(&user->namebuf)[max-1]==FILECACHE_SPECIAL_CHAR)
    max--;
  filecache_advance(max, user);
  if (tino_file_mkdir(tino_buf_get_s(&user->nameout)))
    tino_exit("corrupt_cache, cannot create directory: %s", tino_buf_get_s(&user->nameout));
}

static void
filecache_rearrange(TINO_SLIST list, struct filecache_user *user)
{
#if 0
  tino_exit("rearrange not yet implemented");
#else
  filecache_new(user);
#endif
}

static int
filecache(const char *dir, const char *name, int readonly)
{
  static struct filecache_user	user;
  TINO_SLIST			list;

  filecache_mangle(name, &user);

  /* Special case: No directory.  Just print out the mangled name
   * without path translation.
   */
  if (!*dir)
    {
      out(&user.namebuf);
      return 0;
    }

  tino_buf_reset(&user.nameout);
  tino_buf_add_s(&user.nameout, dir);

  /* make sure user.namebuf is 0-terminated	*/
  for (; *tino_buf_get_s(&user.namebuf); tino_slist_destroy(list))
    {
      int		found;

      deb("reading dir %s", tino_buf_get_s(&user.nameout));
      list		= tino_dir_read_err(tino_buf_get_s(&user.nameout));
      if (!list)
	tino_exit("corrupt cache, directory read error: %s", tino_buf_get_s(&user.nameout));
      found		= tino_slist_iterate_0(list, filecache_pathpart, &user);
      if (found)
	filecache_advance(found, &user);
      else if (tino_glist_count(list)>=maxdir)
	filecache_rearrange(list, &user);
      else
	filecache_new(&user);
    }
  out(&user.nameout);
  return 0;
}

static void
range_check(int nr, int min, int max, const char *what)
{
  if (max<min+min)
    max	= min+min;
  if (nr<min || nr>max)
    tino_exit("%s out of bounds: %d (%d..%d)", what, nr, min, max);
}

int
main(int argc, char **argv)
{
  int		argn, ret, emptydir;
  const char	*s;
  TINO_BUF	buf;

  argn	= tino_getopt(argc, argv, 1, 2,
		      TINO_GETOPT_VERSION(FILECACHE_VERSION)
		      " cachedirectory name..\n"
		      "	Prints a suitable path to store name into the directory.\n"
		      "	If name is missing then filenames are read from stdin."
		      ,

		      TINO_GETOPT_HELP
		      "h	This help"
		      ,

		      TINO_GETOPT_FLAG
		      "b	Default blanks as line terminator"
		      , &null_termchar,

		      TINO_GETOPT_FLAG
		      "d	Debug mode"
		      , &debug_mode,

		      TINO_GETOPT_FLAG
		      "e	Allow empty directory '' (only mangles name)"
		      , &emptydir,

		      TINO_GETOPT_FLAG
		      "f	read (filter) filenames from stdin"
		      , &readfiles,
#if 0
		      TINO_GETOPT_FLAG
		      "i	ignore character case (works only for ASCII)"
		      , &ignorecase,
#endif
		      TINO_GETOPT_INT
		      TINO_GETOPT_DEFAULT
		      "m	max entries in a directory"
		      , &maxdir,
		      200,

		      TINO_GETOPT_INT
		      TINO_GETOPT_DEFAULT
		      "n	max path length where new directory is started"
		      , &maxpath,
		      200,
		      
		      TINO_GETOPT_CHAR
		      "o code	output line termination character\n"
		      "		(default LF or NUL for option -z)"
		      , &out_termchar,

		      TINO_GETOPT_FLAG
		      "r	read only mode, do not create or reorder entries"
		      , &readonly,

		      TINO_GETOPT_CHAR
		      "t code	line termination character (NUL always is one)\n"
		      "		(default LF of not option -z or blanks for option -b)"
		      , &in_termchar,
#if 0
		      TINO_GETOPT_FLAG
		      "u	UTF-8 filenames"
		      , &utf8,
#endif
		      TINO_GETOPT_FLAG
		      "z	Allow NUL as line terminator (changes defaults)"
		      , &null_termchar,

		      NULL
		      );

  if (argn<=0)
    return 1;

  if (readfiles && argc>argn+1)
    tino_exit("name argument not allowed on option -f");
  if (!readfiles && argc<=argn+1)
    tino_exit("missing argument, for read from stdin use option -f");

  if (argv[argn][0])
    {
      if (tino_file_notdir(argv[argn]))
	tino_exit("directory does not exist: %s", argv[argn]);

      range_check(maxpath, 8, pathconf(argv[argn], _PC_NAME_MAX),    "max path length");
      range_check(maxdir,  8, pathconf(argv[argn], _PC_LINK_MAX)-50, "max directory entries");
    }
  else if (!emptydir)
    tino_exit("directory empty but option -e missing");

  filecache_init();

  if (!readfiles)
    return filecache(argv[argn], argv[argn+1], readonly);

  ret	= 0;
  tino_buf_init(&buf);
  while ((s=tino_buf_line_read(&buf, 0, in_termchar ? in_termchar : blanks_termchar ? -1 : null_termchar ? 0 : 10))!=0)
    if (filecache(argv[argn], s, readonly))
      ret	= 1;
  return ret;
}
